const std = @import("std");
const assert = std.debug.assert;
const c = @cImport({
    @cInclude("MiniFB.h");
    @cInclude("MiniFB_enums.h");
});
const testing = std.testing;

pub const Rgb = packed struct {
    r: u8,
    g: u8,
    b: u8,
    _reserved: u8 = 0,
};

pub const KeyMod = packed struct {
    shift: bool = false,
    control: bool = false,
    alt: bool = false,
    super: bool = false,
    capsLock: bool = false,
    numLock: bool = false,
    _reserved: u26 = 0,
};

pub fn Window(comptime TUserData: type) type {
    return packed struct {
        cwin: *c.mfb_window,

        pub const UpdateError = error{
            InvalidWindow,
            InvalidBuffer,
            InternalError,
        };

        pub const State = enum { ok, exit };

        pub const OpenFlags = packed struct {
            resizable: bool = false,
            fullscreen: bool = false,
            fullscreenDesktop: bool = false,
            borderless: bool = false,
            alwaysOnTop: bool = false,
            reserved: u27 = 0,
        };

        pub fn open(title: [*:0]const u8, width: u32, height: u32, flags: OpenFlags) !Window(TUserData) {
            const intFlags = @bitCast(u32, flags);
            const cTitle = @as([*c]const u8, title);
            const cwin: ?*c.mfb_window = c.mfb_open_ex(cTitle, width, height, intFlags);
            if (cwin) |value| {
                const win = Window(TUserData){ .cwin = value };
                assert(@bitCast(usize, win) == @ptrToInt(win.cwin));
                return win;
            } else {
                return error.ItBroke;
            }
        }

        pub fn waitSync(self: Window(TUserData)) bool {
            return c.mfb_wait_sync(self.cwin);
        }

        pub fn update(self: Window(TUserData), buffer: []Rgb) UpdateError!State {
            const rawState = c.mfb_update(self.cwin, @ptrCast(*anyopaque, buffer.ptr));
            switch (rawState) {
                c.STATE_OK => return State.ok,
                c.STATE_EXIT => return State.exit,
                c.STATE_INVALID_WINDOW => return UpdateError.InvalidWindow,
                c.STATE_INVALID_BUFFER => return UpdateError.InvalidBuffer,
                else => return UpdateError.InternalError,
            }
        }

        pub fn setUserData(self: Window(TUserData), data: *TUserData) void {
            c.mfb_set_user_data(self.cwin, data);
        }

        pub fn getUserData(self: Window(TUserData)) ?*TUserData {
            var cData = c.mfb_get_user_data(self.cwin);
            return @ptrCast(?*TUserData, @alignCast(@alignOf(?*TUserData), cData));
        }

        pub const ActiveFunc = fn (win: Window(TUserData), isActive: bool) callconv(.C) void;
        pub fn setActiveCallback(self: Window(TUserData), callback: ActiveFunc) void {
            c.mfb_set_active_callback(self.cwin, @ptrCast(c.mfb_active_func, callback));
        }
    };
}

test "user data is null if never set" {
    const win = try Window(u64).open("abc", 100, 100, .{});
    const data = win.getUserData();
    try testing.expectEqual(@as(?*u64, null), data);
}

test "user data is not null if previously set" {
    const win = try Window(u64).open("abc", 100, 100, .{});
    var data: u64 = 42;
    win.setUserData(&data);
    const expected: u64 = 42;
    try testing.expectEqual(expected, win.getUserData().?.*);
}

pub fn getTargetFPS() u32 {
    return c.mfb_get_target_fps();
}

pub fn setTargetFPS(fps: u32) void {
    c.mfb_set_target_fps(fps);
}

// TODO Figure out how to run this once I have Internet access.
// test "set and get target FPS" {
//     const max = 40;
//     var fps: u32 = 30;
//     while (fps < max) {
//         setTargetFPS(fps);
//         try std.testing.expectEqual(fps, getTargetFPS());
//     }
// }

pub const Timer = extern struct {
    ctimer: *c.mfb_timer,

    pub fn init() !Timer {
        const ctimer: ?*c.mfb_timer = c.mfb_timer_create();
        if (ctimer) |value| {
            return Timer{ .ctimer = value };
        } else {
            return error.ItBroke;
        }
    }

    pub fn deinit(self: Timer) void {
        c.mfb_timer_destroy(self.ctimer);
    }

    pub fn reset(self: Timer) void {
        c.mfb_timer_reset(self.ctimer);
    }

    pub fn now(self: Timer) f64 {
        return c.mfb_timer_now(self.ctimer);
    }

    pub fn delta(self: Timer) f64 {
        return c.mfb_timer_delta(self.ctimer);
    }

    pub fn getFrequency() f64 {
        return c.mfb_timer_get_frequency();
    }

    pub fn getResolution() f64 {
        return c.mfb_timer_get_resolution();
    }
};
